#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <sps2lib.h>
#include "sps2util_Godai.h"
#include "DJR_Mixer.h"
#include "DJR_Audiosystem.h"
#include "DJR_FrameMemMngr.h"
#include "DJR_Timers.h"
#include "UsefulMacros.h"
#include "fontAdobe.h"
#include "DJR_PadManager.h"

//defines
#define HALF_MODE_SCRATCH_SPEED ( 2.f )
#define MIXER_EVENT_NUM_FUNCS 4
#define NUM_DEMO_TRACKS 2

// prototypes
Bool MixerCreateNewFragment( void );
Bool MixerCheckEvents( void );
Bool MixerSetupNewStream( int iTable, Bool bNext );
Bool MixerStartTrackSwitch( void );
void MixerProcessStreaming( void );
int MixerEventGetFirstFreeCallbackSlot( MixerEvent eEvent, 
                                        MixerEventCallbackFunc pCallbackFunc );
int MixerEventGetNumCallbacks( MixerEvent eEvent );

// statics
static const char *_pDemoTrackNames[ NUM_DEMO_TRACKS ][ 2 ] = 
{
  { "sample-audio/DeadBuffalos.raw", "sample-audio/MassAppeal.raw" },
  { "sample-audio/shoryuken.raw", "sample-audio/hadoken.raw" }
};

static short *pSampleBuffer[ 2 ];
static SampleState State[ 2 ];
static TrackStreamState StreamData[ 2 ];
static void *pMemFramePtr[ 2 ];
static short m_pFragment[ 512 ];
static MixerScratchMode _eScratchMode = SCRATCH_MODE_HALF;

MixerEventCallbackFunc 
    _MixerEventCallbacks[ MIXER_EVENT_MAX ][ MIXER_EVENT_NUM_FUNCS ];

/*******************************************/
void MixerInit( void )
{
  int iEvent;
  int iFunc;

  pSampleBuffer[ 0 ] = pSampleBuffer[ 1 ] = NULL;
  pMemFramePtr[ 0 ] = pMemFramePtr[ 1 ] = NULL;
  // all we need to do is initialize the state
  State[ 0 ].fNumSamples = State[ 1 ].fNumSamples = 0.f;
  State[ 0 ].fSpeed = State[ 1 ].fSpeed = 1.f;
  State[ 0 ].fCurPos = State[ 1 ].fCurPos = 0.f;
  State[ 0 ].bLoop = State[ 1 ].bLoop = FALSE;
  State[ 0 ].bLoaded = State[ 1 ].bLoaded = FALSE;
  State[ 0 ].bPlaying = State[ 1 ].bPlaying = FALSE;
  State[ 0 ].fInitialBPM = State[ 1 ].fInitialBPM = 0.f;

  // reset all info on streaming
  StreamData[ 0 ].bStreaming = StreamData[ 1 ].bStreaming = FALSE;
  StreamData[ 0 ].iFileHandle = StreamData[ 1 ].iFileHandle = -1;
  StreamData[ 0 ].iFileSize = StreamData[ 1 ].iFileSize = 0;
  StreamData[ 0 ].iBytesRead = StreamData[ 1 ].iBytesRead = 0;
  StreamData[ 0 ].iTrackNum = StreamData[ 1 ].iTrackNum = 0;

  // reset all func pointerz, yo
  for( iEvent = 0; iEvent < MIXER_EVENT_MAX; iEvent++ )
  {
    for( iFunc = 0; iFunc < MIXER_EVENT_NUM_FUNCS; iFunc++ )
    {
      _MixerEventCallbacks[ iEvent ][ iFunc ] = NULL;
    }
  }

  // default to half scratch mode
  _eScratchMode = SCRATCH_MODE_HALF;
}

/*********************************************/
void MixerKill( void )
{
    FrameMemMngrDeallocateAll( STACKTYPE_T1, TRUE );
    FrameMemMngrDeallocateAll( STACKTYPE_T1, FALSE );
    FrameMemMngrDeallocateAll( STACKTYPE_T2, TRUE );
    FrameMemMngrDeallocateAll( STACKTYPE_T2, FALSE );
}

/**********************************************/
Bool MixerLoadSample( char *pFileName, char iTable, 
                      float fSpeed, Bool bLoop, Bool bReadData )
{
    Bool bSuccess = FALSE;
    struct stat FileInfo;
    if( stat( pFileName, &FileInfo) < 0 )
       return FALSE;

    // turntable 2 and 1 use different memory stacks
    StackType eStackType = iTable == 0?
                             STACKTYPE_T1 :
                             STACKTYPE_T2;

    // get a frame so we can free later
    if( State[ iTable ].bLoaded )
    {
        // if there is already a record loaded on this table
        // then we must have a frame to free 
        //SYS_ASSERT( pMemFramePtr[ iTable ] );
        FrameMemMngrDeallocateFrame( eStackType,
                                     TRUE, 
                                     pMemFramePtr[ iTable ] );
    }

    // obtain a new frame so we can deallocate later
    pMemFramePtr[ iTable ] = 
        FrameMemMngrObtainFrame( eStackType, TRUE );

    // actually allocate the memory
    pSampleBuffer[ iTable ] = ( short * ) 
        FrameMemMngrAllocate( eStackType, TRUE, FileInfo.st_size );

    if( pSampleBuffer[ iTable ] )
    {
        // load the data into the buffer we just allocated
        int fh;

        // short circuit eval: no open for you
        if( !bReadData || ( fh = open( pFileName, O_RDONLY ) ) >= 0 )
	{
	    // are we using this function to only fill out the State struct? 
	    if( bReadData )
	    {
	        int iBytesRead;
	        iBytesRead = read( fh, pSampleBuffer[ iTable ], 
                                   FileInfo.st_size );
                close( fh );
            }

            // setup the sample state
            State[ iTable ].fNumSamples = (float)( FileInfo.st_size / 4 );
            State[ iTable ].fSpeed = fSpeed;
            State[ iTable ].fCurPos = 
                fSpeed > 0.f? 0.f : State[ iTable ].fNumSamples - 1.f;
            State[ iTable ].bLoop = bLoop;
            State[ iTable ].bLoaded = TRUE;
            State[ iTable ].bPlaying = FALSE;
            bSuccess = TRUE;
            if( bLoop )
	    {
	        // sample length in seconds
	        float fTmpBPM = State[ iTable ].fNumSamples / 48000;
                // BPM assuming sample is 2 measures (8 beats) long in 4/4
                fTmpBPM = 8 * 60 / fTmpBPM;
	        State[ iTable ].fInitialBPM = fTmpBPM;
	    }
            else
	    {
	        State[ iTable ].fInitialBPM = -1.f;
            }
            
            printf( "\n\nsuccessfully loaded %s (%f bytes) on table %d\n", 
                pFileName, State[ iTable ].fNumSamples * 4.f, iTable );
            fflush( NULL );
	}
    }

    // todo: clean up memory if we failed to load the file
    if( !bSuccess )
    {
        State[ iTable ].bLoaded = FALSE;
        // if we allocated memory but failed to open the file, deallocate
        if( pSampleBuffer[ iTable ] )
	{
	    FrameMemMngrDeallocateFrame( eStackType,
                                         TRUE,
                                         pMemFramePtr[ iTable ] );
            printf( "\nfailed to open file %s", pFileName );
	}
        else
	    printf( "\nsomething went wrong allocating memory for %s", pFileName );
        fflush( NULL );
        return FALSE;
    }
    
    return TRUE;
}

/***********************************************/
void MixerUnloadSample( char iTurntable )
{
  // free the memory frame
  if( pMemFramePtr[ iTurntable ] && State[ iTurntable ].bLoaded )
  {
    // turntable 2 and 1 use different memory stacks
    StackType eStackType = iTurntable == 0? STACKTYPE_T1 : STACKTYPE_T2;
    FrameMemMngrDeallocateFrame( eStackType,
                                 TRUE, 
                                 pMemFramePtr[ iTurntable ] );
  }

  // re-initialize the state
  State[ iTurntable ].fNumSamples = 0.f;
  State[ iTurntable ].fSpeed = 1.f;
  State[ iTurntable ].fCurPos = 0.f;
  State[ iTurntable ].bLoop = FALSE;
  State[ iTurntable ].bLoaded = FALSE;
  State[ iTurntable ].bPlaying = FALSE;
  State[ iTurntable ].fInitialBPM = 0.f;
}

/***********************************************/
Bool MixerTogglePause( char iTurntable )
{
  return State[ iTurntable ].bPlaying = !State[ iTurntable ].bPlaying;
}

/************************************************/
void MixerProcessScratching( void )
{
  int iTable;
  float fStickY;
  float fVelY;

  if( _eScratchMode == SCRATCH_MODE_HALF )
  {
    // look at each table for samples we can scratch
    for( iTable = 0; iTable < 2; iTable++ )
    {
      // for the time being, you cant scratch loops
      if( State[ iTable ].bLoop || !State[ iTable ].bLoaded )    
        continue;

      PadGetAnalogSticks( 0, iTable, NULL, &fStickY, &fVelY );

      // pulling the record back so scratch backwards
      if( fStickY > 0.15f && State[ iTable ].fCurPos > 0.f )
      {
        //State[ iTable ].fSpeed = -HALF_MODE_SCRATCH_SPEED;
        //State[ iTable ].fSpeed += 2.f * ( fVelY / 1.5f );
        // variety is the spice of life
        State[ iTable ].fSpeed -= 1.f / 4.f;
        State[ iTable ].fSpeed -= ( random() % 2000 ) / 10000.f;
        State[ iTable ].bPlaying = TRUE;
      }

      // else just play the record
      else if( fStickY > -0.15f && fStickY < 0.15f )
      {
        State[ iTable ].bPlaying = TRUE;
        State[ iTable ].fSpeed = 1.f;
      }

      //
      else
      {
        State[ iTable ].bPlaying = FALSE;
      }
    }
  }
  
  State[ iTable ].fSpeed = 
    CLIP_VALU( -MAX_PLAYBACK_SPEED, State[iTable].fSpeed, MAX_PLAYBACK_SPEED );
}

/***************************************************************************/
Bool MixerSetupNewStream( int iTable, Bool bNext )
{
  Bool bSuccess = FALSE;
  StreamData[ iTable ].iTrackNum = 
      ( StreamData[ iTable ].iTrackNum + 1 ) % NUM_DEMO_TRACKS;
   
  StreamData[ iTable ].iFileHandle = 
      open( _pDemoTrackNames[ iTable ][ StreamData[ iTable ].iTrackNum ],
            O_RDONLY );

  if( StreamData[ iTable ].iFileHandle >= 0 )
  {
    struct stat FileInfo;
    stat( _pDemoTrackNames[ iTable ][ StreamData[ iTable ].iTrackNum ], 
           &FileInfo );

    StreamData[ iTable ].bStreaming = TRUE;
    StreamData[ iTable ].iFileSize = FileInfo.st_size;
    StreamData[ iTable ].iBytesRead = 0;

    // set up the record State
    // TODO: determine looping not based on which turntable we're on
    MixerLoadSample( _pDemoTrackNames[iTable][StreamData[iTable].iTrackNum], 
                     iTable, 1.f, !iTable, TRUE );

    bSuccess = TRUE;
  }

  return bSuccess;
}

/***********************************************/
Bool MixerStartTrackSwitch( void )
{
  float fPctDone = 0.75f;
  Bool bSwitch = FALSE;

  // left turntable, next track
  if( PadIsButtonPressed( 0, PS2PAD_BUTTON_L1 ) == KEYPRESS_TOLERANCE )
  {
    if( MixerSetupNewStream( 0, TRUE ) )
    {
      bSwitch = TRUE;
      State[ 0 ].fCurPos = 0.f;
    }
  }

  // left turntable, prev track
  else if( PadIsButtonPressed( 0, PS2PAD_BUTTON_L2 ) == KEYPRESS_TOLERANCE )
  {
    if( State[ 0 ].bLoaded )
    {
      fPctDone = State[ 0 ].fCurPos / State[ 0 ].fNumSamples;

      // works like a CD player.  If we hit back later on in a song, it
      // goes to the beginning of the song instead of the previous track
      if( fPctDone > 0.10f && State[ 0 ].bLoop  )
      {
        
      }
      else
      {
        if( MixerSetupNewStream( 0, FALSE ) )
	{
          bSwitch = TRUE;
        }
      }
      State[ 0 ].fCurPos = 0.f;
    } 
  }

  // right turntable, next track
  if( PadIsButtonPressed( 0, PS2PAD_BUTTON_R1 ) == KEYPRESS_TOLERANCE )
  {
    if( MixerSetupNewStream( 1, TRUE ) )
    {
      bSwitch = TRUE;
      State[ 1 ].fCurPos = 0.f;
    }
  }

  // right turntable, prev track
  else if( PadIsButtonPressed( 0, PS2PAD_BUTTON_R2 ) == KEYPRESS_TOLERANCE )
  {
    if( State[ 1 ].bLoaded )
    {    
      fPctDone = State[ 1 ].fCurPos / State[ 1 ].fNumSamples;
    
      if( fPctDone > 0.10f && State[ 1 ].bLoop  )
      {
        
      }
      else
      {
        if( MixerSetupNewStream( 1, FALSE ) )
	{
          bSwitch = TRUE;
	}
      }
      State[ 1 ].fCurPos = 0.f; 
    }
  }

  return bSwitch;
}

/*******************************************************************/
void MixerProcessStreaming( void )
{
  int iTable;

  for( iTable = 0; iTable < 2; iTable++ )
  {
    if( !StreamData[ iTable ].bStreaming )
      continue;

    // stream another 1024 bytes
    int iBytesRead = 0;
    iBytesRead = 
      read( StreamData[ iTable ].iFileHandle,
            &pSampleBuffer[ iTable ][ StreamData[ iTable ].iBytesRead / 2 ],
            1024 );

    StreamData[ iTable ].iBytesRead += iBytesRead;

    // if we read the whole file into memory, no need to stream anymore
    if( StreamData[ iTable ].iBytesRead >= StreamData[ iTable ].iFileSize )
    {
      StreamData[ iTable ].bStreaming = FALSE;
      close( StreamData[ iTable ].iFileHandle );
    }

    State[ iTable ].bPlaying = TRUE;
  }
}

/************************************************/
Bool MixerProcess( Bool bReset, Bool bTimerExpired )
{
  Bool bExit = FALSE;
  // time to make a new fragment???
  // stop the system from blocking

  // do all controller based updates
  if( bTimerExpired )
  {
      // update pad data
      PadUpdatePadData( );
      // update scratching
      MixerProcessScratching( );
      // does the user want to switch tracks???
      MixerStartTrackSwitch( );
      //  my name's doug and I'm outta...  here
      if( PadIsButtonPressed( 0, PS2PAD_BUTTON_START ) )
          bExit = TRUE;
  }

  if( OSSDeviceMaxBytes() >= 1024 )
  {
    // if we're streaming, load a chunk now
    MixerProcessStreaming( );
    MixerCheckEvents( );
    MixerCreateNewFragment( );
  }

  return bExit;
}

/**************************************************/
Bool MixerCreateNewFragment( void )
{
  // no records loaded...  nothing to do
  if( !State[ 0 ].bLoaded && !State[ 1 ].bLoaded )
    return FALSE;  

  // if we only have one record loaded, set some pointers to avoid crashing
  static short iNotLoadedBuf[ 2 ] = { 0, 0 };
  if( !pSampleBuffer[ 0 ] )
      pSampleBuffer[ 0 ] = &iNotLoadedBuf;
  else if( !pSampleBuffer[ 1 ] )
      pSampleBuffer[ 1 ] = &iNotLoadedBuf;

  // fragment size is 1024 and stereo, 
  // so we have 512 bytes of audio to generate
  // each sample is is a signed short so we generate 256 samples
  unsigned int uSamplesPerFrag = GetFragSize( ) / 2 / 2;

  // TODO: what if its hamster style crossfader???
  float fXFader;
  PadGetAnalogSticks( 0, 0, &fXFader, NULL, NULL );
  fXFader = ( fXFader + 1.f ) / 2.f;
  float fVolume[ 2 ] = { 1.f - fXFader, fXFader };


  int iFragBufPtr = 0;

  while( uSamplesPerFrag > 0 )
  {   
    // mix samples.  fCurPos is mul * 2 because its stereo data
    short iSample = ( short )( pSampleBuffer[ 0 ][ 2 * ROUND( State[ 0 ].fCurPos ) ] * fVolume[ 0 ] );
    iSample += ( short )( pSampleBuffer[ 1 ][ 2 * ROUND( State[ 1 ].fCurPos ) ] * fVolume[ 1 ] );

    // send the stereo sample to the DSP
    m_pFragment[ iFragBufPtr++ ] = iSample;
    m_pFragment[ iFragBufPtr++ ] = iSample;

    // dont write two shorts at a time...  write the whole buffer later
    //WriteSampleToFragment( &m_pFragment[ iFragBufPtr - 2 ] );

    // update sample pointers if the sample isnt paused
    // TODO: play around with rewriting this in assembly 
    int uTable;
    for( uTable=0; uTable<2; uTable++ )
    {
      // if this turntable has a song loaded and its not paused...
      if( State[ uTable ].bLoaded && State[ uTable ].bPlaying )
      {
	State[ uTable ].fCurPos += State[ uTable ].fSpeed;
                    
        // if this sample is supposed to loop
        if( State[ uTable ].bLoop )
	{
	  // going backwards and now at the beginning, so jump to the end
	  if( State[ uTable ].fCurPos <= 0.f )
	    State[ uTable ].fCurPos = State[ uTable ].fNumSamples - 1.f;
                        
          // going forward and now at the end, so go to the beginning
	  else if( State[ uTable ].fCurPos>=State[ uTable ].fNumSamples )
	    State[ uTable ].fCurPos = 0.f;
	}
	else
	{
	  // if we're not looping, clamp to the sample size 
	  if( State[ uTable ].fCurPos < 0 )
	  {
              State[ uTable ].fCurPos = 0;
              State[ uTable ].bPlaying = FALSE;
	  }
          else if( State[ uTable ].fCurPos > State[ uTable ].fNumSamples - 1.f )
	  {
              State[ uTable ].fCurPos = State[ uTable ].fNumSamples - 1;
              State[ uTable ].bPlaying = FALSE;
	  }
	}
      }
    }
    --uSamplesPerFrag;  
  }

  WriteBuffToDSP( m_pFragment );

  return FALSE;
}

/**************************************************************/
void MixerRenderDebugText( void )
{
  int iTable;
  for( iTable=0; iTable<2; iTable++ )
  {
    int iXpos = 100 + 300 * iTable; 
    sps2UPrintf( 0, "\376(%i,10)turntable %d", iXpos, iTable );
    sps2UPrintf( 0, "\376(%i,30)track name: %s", iXpos, "none" );
    sps2UPrintf( 0, "\376(%i,50)speed: %f", iXpos, State[iTable].fSpeed );
    sps2UPrintf( 0, "\376(%i,70)loop: %s", 
        iXpos,
	State[ iTable ].bLoop? "yes" : "no" );
    sps2UPrintf( 0, "\376(%i,90)playing: %s",
        iXpos,
        State[ iTable ].bPlaying? "yes" : "no" );
    sps2UPrintf( 0, "\376(%i,110)position: %ld / %ld",
	iXpos,
	( long int )State[ iTable ].fCurPos, 
	( long int )State[ iTable ].fNumSamples );
    if( State[ iTable ].bLoop )
    {
        sps2UPrintf( 0, "\376(%i,130)approx loop BPM: %f", 
            iXpos, State[ iTable ].fInitialBPM * State[ iTable ].fSpeed );
    }
    else
    {
        sps2UPrintf( 0, "\376(%i,130)approx loop BPM: not a loop", iXpos );
    }
  }
  sps2UPrintfRender( );
}

/*******************************************************************/
float MixerGetRecordSpeed( char iTurntable )
{
  // this is the speed multiplier, NOT the BPM

  // if the record isnt loaded or isnt playing, ignore the speed
  // TODO: return 0 for non-looping records at the end
  if( !State[ iTurntable ].bLoaded ||
      !State[ iTurntable ].bPlaying )
  {
      return( 0.f );
  }

  else
      return( State[ iTurntable ].fSpeed );
}

/************************************************************************/
Bool MixerCheckEvents( void )
{ 
    Bool bEventOccured = FALSE;
    int iMixerEvent;
    int iTable;

    // look for events on both turntables
    for( iTable=0; iTable<2; iTable++ )
    {
        // no events for unloaded records, or ones not currently playing
        if( !State[ iTable ].bLoaded || !State[ iTable ].bPlaying )
	    continue;

        // look through all events to see what happened
        for( iMixerEvent=0; iMixerEvent<MIXER_EVENT_MAX; iMixerEvent++ )
        {
            // no need to check this event if there is no callback for it
            if( MixerEventGetNumCallbacks( iMixerEvent ) == 0 )
                continue;

            // do we want to call this event's callback function?
            Bool bDoCallback = FALSE;

            // get the current and next sample ptr just to save code space 
            float fCur = State[ iTable ].fCurPos;
            float fNxt = fCur + 1024 / 4 * State[ iTable ].fSpeed;
            float fTgt;

            // look for each event and see if it occured
            switch( iMixerEvent )
            {
                // another 1/4th of the loop has played
                case MIXER_EVENT_LOOP_QUARTER:
                {
		    int iTgtLoop;

                    // no loop events for non-looping samples
                    if( !State[ iTable ].bLoop )
		        break;

                    for( iTgtLoop=0; iTgtLoop<=4; iTgtLoop++ )
		    {
		        // target is 0/4, 1/4, 2/4, 3/4, 4/4 of NumSanples 
		        fTgt = iTgtLoop / 4.f;
                        fTgt *= State[ iTable ].fNumSamples;
                        if( BETWEEN( fCur, fTgt, fNxt ) || 
                            BETWEEN( fNxt, fTgt, fCur ) )
	                {
                            bEventOccured = TRUE;
                            bDoCallback = TRUE;
                            break;
                        }
		    }     
                }
                break;

                default:
                    // invalid event
                    break;
            }  

            // if we wanna do the callback, do it now!  get in the chopper!
            if( bDoCallback )
	    {
	      int iFuncNum;
              for( iFuncNum = 0; iFuncNum < MIXER_EVENT_NUM_FUNCS; iFuncNum++ )
              {
                // look through all slots for valid callbacks
                if( _MixerEventCallbacks[ iMixerEvent ][ iFuncNum ] )
                  _MixerEventCallbacks[iMixerEvent][iFuncNum]( iTable, NULL );
              }
	    }
        }
    }   

    return bEventOccured;
}

/*********************************************************************/
int MixerEventGetFirstFreeCallbackSlot( MixerEvent eEvent, 
                                        MixerEventCallbackFunc pCallbackFunc )
{
  int iLoop;

  // look for the first free slot
  for( iLoop = 0; iLoop < MIXER_EVENT_NUM_FUNCS; iLoop++ )
  {
    // unused or already has the target function
    if( !_MixerEventCallbacks[ eEvent ][ iLoop ] ||
      ( _MixerEventCallbacks[ eEvent ][ iLoop ] == pCallbackFunc) )
      break;
  }

  return iLoop;
}

/*********************************************************************/
int MixerEventGetNumCallbacks( MixerEvent eEvent )
{
  int iCallbacks = 0;
  int iLoop;

  // count the number of callback functions this event has
  for( iLoop = 0; iLoop < MIXER_EVENT_NUM_FUNCS; iLoop++ )
  {
    if( _MixerEventCallbacks[ eEvent ][ iLoop ] )
      ++iCallbacks;
  }

  return iCallbacks;
}

/*********************************************************************/
void MixerEventAddCallback( MixerEvent eEvent,
                            MixerEventCallbackFunc pCallbackFunc )
{
  int iNextAvail = MixerEventGetFirstFreeCallbackSlot( eEvent, pCallbackFunc );

  // we already added the max number of callbacks for this event
  if( iNextAvail == MIXER_EVENT_NUM_FUNCS )
  {
    printf( "\n\nfailed to add callback for event %d", eEvent );
    return;  
  }

  // TODO: set the callback for ALL events
  if( eEvent == MIXER_EVENT_ALL )
  {
    //    int iEventCnt;
    //    for( iEventCnt=0; iEventCnt<MIXER_EVENT_MAX; iEventCnt++ )
    //      MixerEventCallbacks[ iEventCnt ] = pCallbackFunc;
  }
  else
  {
    _MixerEventCallbacks[ eEvent ][ iNextAvail ] = pCallbackFunc;
  }

  printf( "\n\nadding callback for event %d slot %d (func = 0x%X)",
      eEvent, iNextAvail, pCallbackFunc );
}

/**********************************************************************/
void MixerEventRemoveCallback( MixerEvent eEvent,
                               MixerEventCallbackFunc pCallbackFunc )
{
  int iLoop;

  // look for the func in the list
  for( iLoop = 0; iLoop < MIXER_EVENT_NUM_FUNCS; iLoop++ )
  {
    // unused or already has the target function
    if( _MixerEventCallbacks[ eEvent ][ iLoop ] == pCallbackFunc )
    {  
      _MixerEventCallbacks[ eEvent ][ iLoop ] = NULL;
      break;
    }
  }
}
