#include <math.h>
#include <stdio.h>
#include <sps2tags.h>
#include <sps2regstructs.h>
#include "DJR_SimpleTexMngr.h"
#include "DJR_Model.h"
#include "gsregs.h"
#include "UsefulMacros.h"
  
static sps2Memory_t **_pMem = NULL;

// *******************************
void InitModel( StaticModel_t *pModel, int iSPS2Descriptor )
{
  pModel->m_bLoaded = FALSE;
  pModel->m_iNumPages = 0;
  pModel->m_iTexSlot = -1;
  pModel->m_iSPS2Descriptor = iSPS2Descriptor;
  pModel->m_pDMAChainPages = NULL;	
}

// *******************************
Bool LoadDMAFile( StaticModel_t *pModel, char *pFileName, int iTexSlot,
                  char *pLogFile )
{
  Bool bSuccess = FALSE;
  struct stat FileInfo;
  int fh = -1;
  FILE *fp = NULL;

  if( pLogFile )
    fp = fopen( pLogFile, "w" );
	
  // get the file size
  if( stat( pFileName, &FileInfo) < 0 )
    return FALSE;
	
  // open the file 
  if( ( fh = open( pFileName, O_RDONLY ) ) == -1 )
    return FALSE;
	
  printf( "\n\nloading static model %s (%f QW)", 
    pFileName, FileInfo.st_size / 16.f );
	
  // make sure we dont already have a mode loaded
  DumpModel( pModel );
	
  // figure out how many pages (4096) we need to store this model's chain
  // not including any needed dma tags except for the first one
	
  // dma tag ( 8 bytes ) + giftag ( 16 bytes )+ texflush ( 16 bytes ) + 
  // tex0_1 ( 16 bytes ) +
  // prim (16 bytes) + FileInfo.st_size
  pModel->m_iNumPages = 8 + 16 + 16 + 16 + 16 + FileInfo.st_size;
  pModel->m_iNumPages = (int)ceilf( pModel->m_iNumPages / 4096.f );
  // now figure out how many more DMA tags we need for n pages
  pModel->m_iNumPages = ( pModel->m_iNumPages * 8 ) + 8 + 16 + 
                        16 + 16 + 16 + FileInfo.st_size;
  pModel->m_iNumPages = (int)ceilf( pModel->m_iNumPages / 4096.f );
  
  if( pLogFile )
    fprintf( fp, "\nusing %d 4096 byte pages\n", pModel->m_iNumPages );

  // allocate some page pointers and then the pages themselves
  pModel->m_pDMAChainPages = ( sps2Memory_t ** )
                                 malloc( pModel->m_iNumPages * 
                                         sizeof( sps2Memory_t * ) );
  _pMem = pModel->m_pDMAChainPages;

  // allocate all the needed pages
  int iPageCnt;
  for( iPageCnt = 0; iPageCnt<pModel->m_iNumPages; iPageCnt++ )
  {
    pModel->m_pDMAChainPages[ iPageCnt ] = sps2Allocate( 4096, 
                                     SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED,
				     pModel->m_iSPS2Descriptor );    	
  }

  // load the first page and insert the initial dmatag, 
  // giftag and registers
  sps2DMATag_t *pDMATag = ( sps2DMATag_t * )
                              ( pModel->m_pDMAChainPages[ 0 ]->pvStart );
  pDMATag->s.PCE = 0; 
  pDMATag->s.SPR = 0;   // main mem, not scratchpad
  pDMATag->s.IRQ = 0;    // no interrupt

  if( pModel->m_iNumPages > 1 )
  {
    pDMATag->s.QWC = 255;
    pDMATag->s.ID = DMA_ID_SOURCE_NEXT;
    pDMATag->s.ADDR = 
      ( sps2uint32 )
        sps2GetPhysicalAddress( pModel->m_pDMAChainPages[ 1 ]->pvStart,
                                pModel->m_pDMAChainPages[ 1 ] );
  }
  else
  {
    pDMATag->s.QWC = ( 8 + 16 + 16 + 16 + 16 + FileInfo.st_size ) / 16; 
    pDMATag->s.ID = DMA_ID_SOURCE_END;
    pDMATag->s.ADDR = 0;
  }

  if( pLogFile )
    LogModelTag( fp, MODEL_DMATAG, pDMATag, pModel->m_pDMAChainPages[0] );

  pDMATag++;

  sps2GIFTag_t *gt = (sps2GIFTag_t *)pDMATag;
  sps2uint64 *regs = (sps2uint64 *) ((char*)gt+sizeof(sps2GIFTag_t));

  gt->i128 = 0;  
  gt->s.NLOOP = 1; 
  gt->s.EOP = 0;
  gt->s.PRE = 0;   
  gt->s.PRIM = 0;
  gt->s.FLG = GIF_FLG_REGLIST;
  gt->s.NREG = 0;

  // add an A_D register
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // TEXFLUSH
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // TEX0_1 
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // PRIM

  if( pLogFile )
    LogModelTag( fp, MODEL_GIFTAG, gt, pModel->m_pDMAChainPages[ 0 ] );

  // access the texflush register with anything
  *(regs++) = (sps2uint64)0xAA55;
  *(regs++) = TEXFLUSH_R;
  sps2uint32 tex_mem_addr = GetSlotGSAddress( iTexSlot );
  *(regs++) = PACK_TEX0(tex_mem_addr/64,TEX_W/64,0,8,8,1,1,0,0,0,0,0);
  *(regs++) = TEX0_1_R;
  *(regs++) = PACK_PRIM(3,1,1,0,0,1,0,0,0);
  *(regs++) = PRIM_R;

  // try to read enough to fill the page
  // on the first page, 80 bytes are used for dma tags, 
  // the gif tag, and regs
  int iBytesRead = read( fh, regs, 4096 - 80 );
	
  // load the rest of the pages
  for( iPageCnt = 1; iPageCnt<pModel->m_iNumPages; iPageCnt++ )
  {
    pDMATag = ( sps2DMATag_t * )
                ( pModel->m_pDMAChainPages[ iPageCnt ]->pvStart );
    if( iPageCnt < pModel->m_iNumPages - 1 )
    {
      pDMATag->s.ID = DMA_ID_SOURCE_NEXT;
      pDMATag->s.ADDR = 
      ( sps2uint32 )
      sps2GetPhysicalAddress(pModel->m_pDMAChainPages[ iPageCnt + 1 ]->pvStart,
                             pModel->m_pDMAChainPages[ iPageCnt + 1 ] );
    }
    else
    {
      pDMATag->s.ID = DMA_ID_SOURCE_END;
      pDMATag->s.ADDR = 0;
    }
    iBytesRead = read( fh, ( void * )( pDMATag + 1 ), 4096 - 16 );
    pDMATag->s.QWC = iBytesRead / 16;
  }
	
  fclose( fp );
  close( fh );

  return TRUE;
}

// *******************************
void DumpModel( StaticModel_t *pModel )
{
  pModel->m_iTexSlot = -1;

  if( pModel->m_iNumPages > 0 && pModel->m_pDMAChainPages )
  {
    while( --( pModel->m_iNumPages ) >= 0 )
    {
      sps2Free( pModel->m_pDMAChainPages[ pModel->m_iNumPages ] );
    }
	
    free( pModel->m_pDMAChainPages );
    pModel->m_pDMAChainPages = NULL;	
  }
  pModel->m_bLoaded = FALSE;
}

// *******************************
void RenderModel( StaticModel_t *pModel )
{
  Dn_CHCR_t chcrValue;
  Dn_TADR_t tadrValue;

  chcrValue.i32 = 0;
  chcrValue.s.DIR = 1;
  chcrValue.s.MOD = 1;
  chcrValue.s.STR = 1;

  *EE_D2_QWC= 0;
  tadrValue.i32 = 0;
  tadrValue.s.ADDR = 
    sps2GetPhysicalAddress( pModel->m_pDMAChainPages[ 0 ]->pvStart,
                            pModel->m_pDMAChainPages[ 0 ] );
  *EE_D2_TADR = tadrValue.i32;
  *EE_D2_CHCR = chcrValue.i32;  	
}

/*********************************************************************/
Bool LogModelTag( FILE *fp, eModelTag tagtype, 
                  void *pTag, sps2Memory_t *pPage )
{
  switch( tagtype )
  {
    case MODEL_DMATAG:
    {
      const char *pTagTypes[8] = { "refe", "cnt", "next", "ref", 
                                 "refs", "call", "ret", "end" };
      sps2DMATag_t *pDMATag = ( sps2DMATag_t * )pTag;
      fprintf( fp, "\n\nDMA tag at addr 0x%X ( physical = 0x%X )\n",
        pTag, sps2GetPhysicalAddress( pTag, pPage ) );
      fprintf(fp,"QWC: %d\nPCE: %d\nID: %s\nIRQ: %d\nADDR: 0x%X\nSPR: %d", 
          pDMATag->s.QWC, pDMATag->s.PCE, pTagTypes[pDMATag->s.ID], 
          pDMATag->s.IRQ, pDMATag->s.ADDR, pDMATag->s.SPR );
    }
    break;

    case MODEL_GIFTAG:
    {
      const char *pDataFormat[ 4 ] = { "PACKED", "REGLIST", 
                                       "IMAGE", "DISABLED" };
      sps2GIFTag_t *pGT = ( sps2GIFTag_t * )pTag;
      fprintf( fp, "\n\nGIF tag at addr 0x%X (physical = 0x%X )\n",
        pTag, sps2GetPhysicalAddress( pTag, pPage ) );
      fprintf(fp,"NLOOP: %d\nEOP: %d\nPRE: %d\nPRIM: %d\nFLG: %s\nNREG: %d\nREGS: 0x%X",
        pGT->s.NLOOP, pGT->s.EOP, pGT->s.PRE, pGT->s.PRIM,
        pDataFormat[ pGT->s.FLG ], pGT->s.NREG, pGT->s.REGS );
    }
    break;

    default:
      break;
  }

  return FALSE;
}
