/*
    Copyright (C) 2002 Terratron Technologies Inc.  All Rights Reserved.
    Author: Steven Osman, Morten Mikkelsen, Lionel Lemarie
 
    This file is part of sps2.

    Before using this file you MUST agree to the license agreement in
    the file LICENSE provided with this package.
  
    For more information visit:
    http://www.playstation2-linux.com/projects/sps2/
    http://window.terratron.com/~sosman/ps2linux/
    http://www.terratron.com/

    7/12/2003 - Sauce
    Initial version

    8/8/2003 - Sauce
    Split sps2UScreenSwap to: sps2UScreenSwap and sps2UScreenClear(r, g, b)
*/
#ifndef __SPS2_UTIL_H
#define __SPS2_UTIL_H

#include "sps2types.h"
#include "sps2lib.h"

#ifdef __cplusplus
extern "C" {
#endif

#define SPS2U_SCR_VESA		0
#define SPS2U_SCR_NTSC		2
#define SPS2U_SCR_PAL		3

  /**
   * Creates a virtual console, builds a double and Z buffer and prepares the
   * GS for graphics output.
   * iSuppressSignalHandlers -- if zero, then sps2UScreenInit will trap all
   * signals and make sure that the display is shutdown gracefully before the
   * program aborts when it receives a signal.  If set to non-zero, then
   * the programmer is responsible for making sure that the screen is properly
   * shut down when the program aborts.
   */ 
  void sps2UScreenInit(int iSuppressSignalHandlers);

  /**
   * Destroys the virtual console and reverts the GS to it's original state.
   * You should make sure to call this after calling sps2UScreenInit.
   */
  void sps2UScreenShutdown();

  /**
   * VSync and swap the two buffers.  The one that was being drawn will now
   * be displayed and the one that was being displayed is prepared for drawing.
   * It is not, however cleared.  Use sps2UScreenClear() to clear the new
   * drawing screen.
   */
  void sps2UScreenSwap();
  void sps2UScreenSwap_NoVsyncWait();

  /**
   * Clears the drawing screen with the specified color.
   */
  void sps2UScreenClear(int iRed, int iGreen, int iBlue);

  /**
   * Returns the page number of the first available page in the GS memory,
   * after the two buffers and Z buffer.
   */
  unsigned int sps2UScreenGetFirstFreeGSPage();

  /**
   * Returns the first page number of the Z buffer
   */
  unsigned int sps2UScreenGetZPtr();

  /**
   * Returns the first page number of the first draw buffer.
   */
  unsigned int sps2UScreenGetDrawBuff1();

  /**
   * Returns the first page number of the second draw buffer.
   */
  unsigned int sps2UScreenGetDrawBuff2();

  /**
   * Returns the first page number of the current buffer that is being drawn
   * to.
   */
  unsigned int sps2UScreenGetDrawBuffCurrent();

  /**
   * Returns the width of the current screen resolution.
   * Note, the current implementation of sps2UScreen always sets a width of
   * 640 pixels.
   */
  unsigned int sps2UScreenGetWidth();

  /**
   * Returns the height of the current screen resolution.
   * Note, the current implementation of sps2UScreen always sets a height of:
   * 480 for sps2UScreenGetVideoMode() == SPS2U_SCR_VESA
   * 448 for sps2UScreenGetVideoMode() == SPS2U_SCR_NTSC
   * 512 for sps2UScreenGetVideoMode() == SPS2U_SCR_PAL
   */
  unsigned int sps2UScreenGetHeight();

  /**
   * Returns the Z depth of the screen configuration (i.e. the largest possible
   * Z value).
   * Note, since this is currently always 24 bit, this function will always
   * return 0xffffff.
   */
  unsigned int sps2UScreenGetZDepth();

  /**
   * Returns the video mode that has been auto detected.  The choices are:
   *
   * SPS2U_SCR_VESA
   * SPS2U_SCR_NTSC
   * SPS2U_SCR_PAL
   */
  unsigned int sps2UScreenGetVideoMode();
  
  /**
   * Returns the pixel format of the current display settings.
   * Note, the current implementation of sps2UScreen always returns a
   * pixel format of 0 which corresponds to PSMCT32.
   */
  unsigned int sps2UScreenGetPixelFormat();

  /**
   * Called _before_ sps2UScreenInit, this disables the use of gsvnc.
   * Note that gsvnc offers only a small memory overhead and no performance
   * overhead if you don't have a vnc server running _and_ a client connected
   * to it, so generally you'd want to leave the support in there.
   */
  void sps2UScreenDisableVNC();

  /**
   * Sets the frequency at which the vnc server should get an updated frame.
   * Note that frame grabs only happen if there's a client connected to the
   * vnc server.
   */
  void sps2UScreenSetVNCUpdateRate(int iUpdateRate);

  /**
   * The code that follows is to support the sps2UPrintf() function.
   * 
   * Usage:
   * sps2UPrintfLoadFont(0,fbp);
   * sps2UPrintfSetWindow(0,0,320,200);
   *
   * foreach frame do {
   *     sps2UPrintf(0,"Hello, world!\nHow are you today");
   *     sps2UPrintf(0,"The next word is in \377FF0000RED!\n");
   *     sps2UPrintf(0,"\376(0,0)Top-o-the-screen.\n");
   *     .
   *     .
   *     .
   *     sps2UPrintfRender();
   * }
   */

#ifndef __SPS2_U_FONT_STRUCT
#define __SPS2_U_FONT_STRUCT
  
  typedef struct {
    int iFirstCharacter;
    int iLastCharacter;
    int iAscent;
    int iDescent;
	unsigned int *auiCharacters[256];
	int aiNonZeroBits[256];
	int aiWidths[256];
} sps2UFontStruct;
#endif

  /**
   * Print some text using font pFont.  If pFont = null, it'll default to
   * a fixed font.
   * You can embed \377RRGGBB into your text to switch colors of the
   * text.
   * You can also embed \376(x, y) to change the coordinates relative to
   * the window you've defined with sps2UPrintfSetWindow().  This is the
   * same as calling sps2UPrintfSetPos()
   * Syntax follows printf syntax.
   */
  int sps2UPrintf(sps2UFontStruct *pFont, const char *format, ...);

  /**
   * Renders the currently buffered text and resets the position to the top
   * of the print window.
   */
  void sps2UPrintfRender();

  /**
   * Sets the window that the text is output to.  Default is (x,y) at the top
   * and (iWidth, iHeight) to span the whole screen.
   * Note that changes to sps2UPrintfSetWindow only apply to the next set
   * of printfs, not the ones already in the queue.
   */
  void sps2UPrintfSetWindow(int iX, int iY, int iWidth, int iHeight);

  /**
   * Returns the window that the text is output to.
   */
  void sps2UPrintfGetWindow(int *piX, int *piY, int *piWidth, int *piHeight);

  /**
   * Sets the current (x,y) coordinates relative to the top of the printing
   * window.  This is the same as embedding \376(iX, iY) to the front of your
   * next string.
   */
  void sps2UPrintfSetPos(int iX, int iY);

  /**
   * Returns the current (x,y) coordinates relative to the top of the printing
   * window.
   */
  void sps2UPrintfGetPos(int *piX, int *piY);

  /**
   * Sets the Z value for printing.  Defaults to 128.
   * Note that changes to sps2UPrintfSetZ only apply to the next set
   * of printfs, not the ones already in the queue.
   */
  void sps2UPrintfSetZ(int iZ);

  /**
   * Returns the Z value being used for the text.
   */
  int sps2UPrintfGetZ();

  /**
   * This function returns the height of the specified font.
   */
  int sps2UPrintfFontHeight(sps2UFontStruct *pFont);

  /**
   * Computes and returns the width of a given string for a specific font.
   * If pFont == 0 then the default font is used
   */
  int sps2UPrintfStringWidth(sps2UFontStruct *pFont, const char *pcString);

  /**
   * The following sets of functions are non-inline versions of their
   * counterparts in sps2lib.h.  For documentation of any of the specific
   * functions, please see sps2lib.h or refer to the manual.
   * These functions were created to assist those who want to use sps2
   * outside of a C programming environment.  For example, an assembly
   * program could call sps2NIInit() and link to libsps2util.a though it
   * would not have been able to expand the inline version defined in
   * sps2lib.h
   */
  void *_sps2UNIMapEERegisters(void *pvWhere, int iDeviceHandle);
  void *_sps2UNIMapGSRegisters(void *pvWhere, int iDeviceHandle);
  void *_sps2UNIMapVUMemory(void *pvWhere, int iDeviceHandle);
  void *_sps2UNIMapScratchPad(void *pvWhere, int iDeviceHandle);
  int sps2UNIInit();
  void sps2UNIRelease(int iDevice);
  int _sps2UNIOpen();
  void _sps2UNIClose(int iDeviceHandle);
  int sps2UNIFlushCache(int iDeviceHandle);
  int _sps2UNIEnableCOP2Access(int iDeviceHandle);
  unsigned long sps2UNIGetPhysicalAddress(void *pvAddress,
					  sps2Memory_t *pDescriptor);
  sps2Memory_t *sps2UNIAllocate(unsigned long ulSize,
				int iMapOptions,
				int iDeviceHandle);
  sps2Memory_t *sps2UNIRemap(sps2Memory_t *pOriginalArea,
			     int iMapOptions,
			     int iDeviceHandle);
  void sps2UNIFree(sps2Memory_t *pMapping);
  int sps2UNIWaitForDMA(int iChannel,
			int iDeviceHandle);
  int sps2UNISetOperationMode(int iSupervisorMode,
			      int iDeviceHandle);
  int _sps2UNISetEIDIEnabled(int iEnabled,
			     int iDeviceHandle);

  /*
    int sps2UNIValidateDMA(int iDeviceHandle, int iChannel,
    sps2uint32 ui32CHCR, int iSizeLimit);
  */
#ifdef __cplusplus
}
#endif

#endif
